"""Routines for handling strings."""

from math import isnan, isfinite, inf
from re import sub
from typing import Union, Optional, Final, Iterable, List, cast, Callable

from moptipy.utils.math import __try_int, try_int
from moptipy.utils.types import type_error


def float_to_str(x: float) -> str:
    """
    Convert `float` to a string.

    :param x: the floating point value
    :return: the string representation

    >>> float_to_str(1.3)
    '1.3'
    >>> float_to_str(1.0)
    '1'
    """
    if x == 0:
        return "0"
    s = repr(x)
    if isnan(x):
        raise ValueError(f"'{s}' not permitted.")
    if s.endswith(".0"):
        return s[:-2]
    return s


def num_to_str_for_name(x: Union[int, float]) -> str:
    """
    Convert a float to a string for use in a component name.

    :param x: the float
    :returns: the string

    >>> num_to_str_for_name(1.3)
    '1d3'
    >>> num_to_str_for_name(1.0)
    '1'
    >>> num_to_str_for_name(-7)
    'm7'
    >>> num_to_str_for_name(-6.32)
    'm6d32'
    """
    return num_to_str(x).replace('.', 'd').replace('-', 'm')


def bool_to_str(value: bool) -> str:
    """
    Convert a Boolean value to a string.

    :param value: the Boolean value
    :return: the string

    >>> print(bool_to_str(True))
    T
    >>> print(bool_to_str(False))
    F
    """
    return 'T' if value else 'F'


def str_to_bool(value: str) -> bool:
    """
    Convert a string to a boolean value.

    :param value: the string value
    :return: the boolean value

    >>> str_to_bool("T")
    True
    >>> str_to_bool("F")
    False
    >>> try:
    ...     str_to_bool("x")
    ... except ValueError as v:
    ...     print(v)
    Expected 'T' or 'F', but got 'x'.
    """
    if value == "T":
        return True
    if value == "F":
        return False
    raise ValueError(f"Expected 'T' or 'F', but got '{value}'.")


def num_to_str(value: Union[int, float]) -> str:
    """
    Transform a numerical type to a string.

    :param value: the value
    :return: the string

    >>> num_to_str(1)
    '1'
    >>> num_to_str(1.5)
    '1.5'
    """
    return str(value) if isinstance(value, int) else float_to_str(value)


def intfloatnone_to_str(val: Union[int, float, None]) -> str:
    """
    Convert an integer ot float or `None` to a string.

    :param val: the value
    :return: the string representation
    :rtype: str

    >>> print(repr(intfloatnone_to_str(None)))
    ''
    >>> print(intfloatnone_to_str(12))
    12
    >>> print(intfloatnone_to_str(12.3))
    12.3
    """
    return "" if val is None else num_to_str(val)


def intnone_to_str(val: Optional[int]) -> str:
    """
    Convert an integer or `None` to a string.

    :param val: the value
    :return: the string representation

    >>> print(repr(intnone_to_str(None)))
    ''
    >>> print(intnone_to_str(12))
    12
    """
    return "" if val is None else str(val)


def str_to_intfloat(val: str) -> Union[int, float]:
    """
    Convert a string to an int or float.

    :param val: the string value
    :return: the int or float

    >>> print(type(str_to_intfloat("15.0")))
    <class 'int'>
    >>> print(type(str_to_intfloat("15.1")))
    <class 'float'>
    """
    return __try_int(float(val)) if ("e" in val) or \
                                    ("E" in val) or \
                                    ("." in val) or \
                                    ("inf" in val) else int(val)


def str_to_intfloatnone(val: str) -> Union[int, float, None]:
    """
    Convert a string to an int or float or None.

    :param val: the string value
    :return: the int or float or None

    >>> print(str_to_intfloatnone(""))
    None
    >>> print(type(str_to_intfloatnone("5.0")))
    <class 'int'>
    >>> print(type(str_to_intfloatnone("5.1")))
    <class 'float'>
    """
    return None if len(val) <= 0 else str_to_intfloat(val)


def str_to_intnone(val: str) -> Optional[int]:
    """
    Convert a string to an int or None.

    :param val: the string value
    :return: the int or None

    >>> print(str_to_intnone(""))
    None
    >>> print(str_to_intnone("5"))
    5
    """
    return None if len(val) <= 0 else int(val)


def replace_all(find: str, replace: str, src: str) -> str:
    """
    Perform a recursive replacement of strings.

    After applying this function, there will not be any occurence of `find`
    left in `src`. All of them will have been replaced by `replace`. If that
    produces new instances of `find`, these will be replaced as well.
    If `replace` contains `find`, this will lead to an endless loop!

    :param find: the string to find
    :param replace: the string with which it will be replaced
    :param src: the string in which we search
    :return: the string `src`, with all occurrences of find replaced by replace

    >>> replace_all("a", "b", "abc")
    'bbc'
    >>> replace_all("aa", "a", "aaaaa")
    'a'
    >>> replace_all("aba", "a", "abaababa")
    'aa'
    """
    new_len = len(src)
    while True:
        src = src.replace(find, replace)
        old_len = new_len
        new_len = len(src)
        if new_len >= old_len:
            return src


def __replace_double(replace: str, src: str) -> str:
    """
    Replace any double-occurrence of a string with a single occurrence.

    :param replace: the string to replace
    :param src: the source string
    :returns: the updated string
    """
    return replace_all(replace + replace, replace, src)


#: the separator of different filename parts
PART_SEPARATOR: Final[str] = "_"
#: the replacement for "." in a file name
DECIMAL_DOT_REPLACEMENT: Final[str] = "d"


def sanitize_name(name: str) -> str:
    """
    Sanitize a name in such a way that it can be used as path component.

    >>> sanitize_name(" hello world ")
    'hello_world'
    >>> sanitize_name(" 56.6-455 ")
    '56d6-455'
    >>> sanitize_name(" _ i _ am _ funny   --6 _ ")
    'i_am_funny_-6'

    :param name: the name that should be sanitized
    :return: the sanitized name
    :raises ValueError: if the name is invalid or empty
    :raises TypeError: if the name is `None` or not a string
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    orig_name = name
    name = name.strip()
    name = __replace_double("-", name)
    name = __replace_double("_", name)
    name = __replace_double(".", name).replace(".", DECIMAL_DOT_REPLACEMENT)

    name = sub(r"[^\w\s-]", '', name)
    name = sub(r"\s+", PART_SEPARATOR, name)
    name = __replace_double("_", name)

    if name.startswith("_"):
        name = name[1:]

    if name.endswith("_"):
        name = name[:len(name) - 1]

    if len(name) <= 0:
        raise ValueError(
            f"Sanitized name must not become empty, but '{orig_name}' does.")

    return name


def sanitize_names(names: Iterable[str]) -> str:
    """
    Sanitize a set of names.

    >>> sanitize_names(["", " sdf ", "", "5-3"])
    'sdf_5-3'
    >>> sanitize_names([" a ", " b", " c", "", "6", ""])
    'a_b_c_6'

    :param names: the list of names.
    :return: the sanitized name
    """
    return PART_SEPARATOR.join([
        sanitize_name(name) for name in names if len(name) > 0])


#: the internal integer to float conversion threshold
__INT_TO_FLOAT_THRESHOLD: Final[float] = 1E10


def default_float_format(min_finite: Union[int, float] = 0,
                         max_finite: Union[int, float] = 0,
                         frac_len: int = 2) -> str:
    """
    Get the default float format.

    :param min_finite: the minimum finite value
    :param max_finite: the maximum finite value
    :param frac_len: the longest fraction
    """
    if not isinstance(min_finite, (int, float)):
        raise type_error(min_finite, "min_finite", (int, float))
    if not isinstance(max_finite, (int, float)):
        raise type_error(max_finite, "max_finite", (int, float))
    if not (isfinite(min_finite) and isfinite(max_finite)
            and (min_finite <= max_finite)):
        raise ValueError("invalid min_finite, max_finite pair "
                         f"{min_finite}, {max_finite}.")
    if not isinstance(frac_len, int):
        raise type_error(frac_len, "frac_len", int)
    if not (0 <= frac_len < 100):
        raise ValueError(f"invalid frac_len {frac_len}.")

    if ((-__INT_TO_FLOAT_THRESHOLD) <= min_finite) \
            and (max_finite <= __INT_TO_FLOAT_THRESHOLD):
        if (frac_len <= 0) or (min_finite <= -1E4) or (max_finite >= 1E4):
            return "{:.0f}"
        if (frac_len <= 1) or (min_finite <= -1E3) or (max_finite >= 1E3):
            return "{:.1f}"
        if (frac_len <= 2) or (min_finite <= -1E2) or (max_finite >= 1E2):
            return "{:.2f}"
        return "{:.3f}"
    return "{:.2e}"


def numbers_to_strings(source: Union[int, float, None,
                                     Iterable[Union[int, float, None]]],
                       none_str: Optional[str] = None,
                       nan_str: Optional[str] = r"$\emptyset$",
                       positive_infty_str: Optional[str] = r"$\infty$",
                       negative_infty_str: Optional[str] = r"$-\infty$",
                       int_renderer: Optional[Callable[[int], str]] = None,
                       float_format_getter: Callable[
                           [Union[int, float], Union[int, float], int], str]
                       = default_float_format,
                       exponent_renderer: Callable[[str], str] =
                       lambda e: f"*10^{e}^",
                       int_to_float_threshold:
                       Union[int, float] = __INT_TO_FLOAT_THRESHOLD) \
        -> List[Optional[str]]:
    r"""
    Convert a numerical column to text with uniform shape.

    Often, we need to convert a set of numbers to strings as output for a
    table or another representative thext. In such a case, you want to present
    all numbers in the set in the same format.
    Imagine you have the number vector `[1E-4, 1/7, 123456789012345678]`. If
    you simply convert this list to a string directly, what you get is
    `[0.0001, 0.14285714285714285, 123456789012345678]`. Now this looks very
    ugly. First, we have one very big number `123456789012345678`. If the
    numbers stem from an experiment, then we hardly are able to obtain any
    number at a very extreme precision. The 18 digits in `123456789012345678`
    sort of suggest a precision to 18 decimals, since the number ends in
    specific digits (as opposed to `123450000000000000` which a reader would
    naturally preceive as a rounded quantity). Additionally, we the number
    `0.14285714285714285`, which has a very long fractional part, which, too,
    suggests a very high precision. Writing both mentioned numbers next to
    each other, this suggests as if we could present a number as high as
    10**18 at a precision of 10**-17. And it also looks ugly, because both
    numbers are not uniformly formatted. Instead, our function here renders
    the number list as `['1.00*10^-4^', '1.43*10^-1^', '1.23*10^17^']`. It
    recognizes that we should present numbers as powers of ten and then limits
    the precision to three digits.

    This function is thus intended to produce some sort of uniform format with
    reasonable precision uniformly for a numerical vector, under the
    assumption that all numbers should be presented in the same numerical range
    and quantity.

    :param source: the column data
    :param none_str: the string replacement for `None`
    :param nan_str: the string to be used for NaN
    :param positive_infty_str: the string to be used for positive infinity
    :param negative_infty_str: the string to be used for negative infinity
    :param int_renderer: the renderer for integers
    :param float_format_getter: a float format getter
    :param exponent_renderer: the renderer for exponents
    :param int_to_float_threshold: the absolute threshold after which integers
        will be force-converted to floating point numbers to ensure that we do
        not express uselessly large numbers as integers
    :returns: a list with the text representation

    >>> from moptipy.utils.lang import EN
    >>> EN.set_current()
    >>> numbers_to_strings([1.75651, 212, 3234234])
    ['2', '212', "3'234'234"]
    >>> numbers_to_strings([1.75651, 22, 34])
    ['1.757', '22.000', '34.000']
    >>> numbers_to_strings([1.75651, 122, 34])
    ['1.76', '122.00', '34.00']
    >>> numbers_to_strings([1.75651, 122, 3334])
    ['1.8', '122.0', "3'334.0"]
    >>> numbers_to_strings([1.5, 212, 3234234])
    ['2', '212', "3'234'234"]
    >>> numbers_to_strings([1.5, 2e12, 3234234])
    ['1.50*10^0^', '2.00*10^12^', '3.23*10^6^']
    >>> numbers_to_strings([233, 22139283482834, 3234234])
    ['2.33*10^2^', '2.21*10^13^', '3.23*10^6^']
    >>> numbers_to_strings([233, 22139283, 3234234])
    ['233', "22'139'283", "3'234'234"]
    >>> from math import nan, inf
    >>> numbers_to_strings([22139283, inf, -inf, nan, None])
    ["22'139'283", '$\\infty$', '$-\\infty$', '$\\emptyset$', None]
    >>> numbers_to_strings([1E-4, 1/7, 123456789012345678])
    ['1.00*10^-4^', '1.43*10^-1^', '1.23*10^17^']
    """
    # perform type checks
    if (source is None) or isinstance(source, (int, float)):
        source = [source]
    if not isinstance(source, Iterable):
        raise type_error(source, "source", Iterable)
    if (none_str is not None) and (not isinstance(none_str, str)):
        raise type_error(none_str, "none_str", (str, None))
    if (nan_str is not None) and (not isinstance(nan_str, str)):
        raise type_error(nan_str, "nan_str", (str, None))
    if (positive_infty_str is not None) \
            and (not isinstance(positive_infty_str, str)):
        raise type_error(positive_infty_str, "positive_infty_str",
                         (str, None))
    if (negative_infty_str is not None) \
            and (not isinstance(negative_infty_str, str)):
        raise type_error(negative_infty_str, "negative_infty_str",
                         (str, None))
    if int_renderer is None:
        from moptipy.utils.lang import Lang  # pylint: disable=C0415,R0401
        int_renderer = Lang.current().format_int
    if not callable(int_renderer):
        raise type_error(int_renderer, "int_renderer", call=True)
    if not callable(float_format_getter):
        raise type_error(float_format_getter,
                         "float_format_getter", call=True)
    if not callable(exponent_renderer):
        raise type_error(exponent_renderer,
                         "exponent_renderer", call=True)
    if not isinstance(int_to_float_threshold, (int, float)):
        raise type_error(int_to_float_threshold, "int_to_float_threshold",
                         (int, float))
    if isnan(int_to_float_threshold) or (int_to_float_threshold <= 0):
        raise ValueError(
            f"invalid int_to_float_threshold: {int_to_float_threshold}.")

    # step one: get the raw numerical data
    data: Final[List[Union[int, float, None]]] = \
        cast(List, source) if isinstance(source, List) else list(source)
    dlen: Final[int] = len(data)
    if dlen <= 0:
        raise ValueError("Data cannot be empty.")

    # step two: investigate the data ranges and structure
    all_is_none: bool = True
    all_is_int: bool = True
    max_finite: Union[int, float] = -inf
    min_finite: Union[int, float] = inf
    longest_fraction: int = -1

    for i, d in enumerate(data):
        if d is None:
            continue
        all_is_none = False
        d2 = try_int(d) if isfinite(d) else d
        if isinstance(d2, int):
            if d2 < min_finite:
                min_finite = d2
            if d2 > max_finite:
                max_finite = d2
            if not ((-int_to_float_threshold) <= d2
                    <= int_to_float_threshold):
                d2 = float(d2)
        if d2 is not d:
            data[i] = d2

        if isfinite(d2):
            if not isinstance(d2, int):
                all_is_int = False
                s = str(d2)
                if not (("E" in s) or ("e" in s)):
                    i = s.find(".")
                    if i >= 0:
                        i = len(s) - i - 1
                        if i > longest_fraction:
                            longest_fraction = i
            if d2 < min_finite:
                min_finite = d2
            if d2 > max_finite:
                max_finite = d2

    # step three: if all data is None, we can return here
    if all_is_none:
        return [none_str] * dlen

    # create the protected integer renderer
    def __toint(value: int, form=int_renderer) -> str:
        sv: str = form(value).strip()
        if (sv is not None) and (not isinstance(sv, str)):
            raise type_error(s, f"conversion of {value}", (str, None))
        return sv

    # step four: if all data are integer, we can convert them directly
    if all_is_int:
        # an int render also processing None and special floats
        def __toint2(value: Union[None, int, float],
                     form=__toint,
                     na=nan_str,
                     pi=positive_infty_str,
                     ni=negative_infty_str) -> Optional[str]:
            if value is None:
                return None
            return form(cast(int, value)) if isfinite(value)\
                else na if isnan(value) else pi if value >= inf else ni
        return [__toint2(i) for i in data]

    # ok, we have at least some finite floats that cannot be converted to
    # integers. therefore, we need to convert them to strings based on a
    # floating point number format.
    float_format = float_format_getter(min_finite, max_finite,
                                       longest_fraction)
    if not isinstance(float_format, str):
        raise type_error(float_format,
                         "float format from float_format_getter", str)
    if (len(float_format) <= 0) or ('{' not in float_format) \
            or ('}' not in float_format) or (':' not in float_format):
        raise ValueError(f"invalid float format '{float_format}'.")

    # step five: first, create the raw float strings and mark special values
    result: Final[List[Union[int, str]]] = [
        0 if value is None
        else float_format.format(value).strip() if isfinite(value)
        else 1 if isnan(value)
        else 2 if value >= inf
        else 3 for value in data]

    # step six: fix special values and float strings
    for i, value in enumerate(result):
        if value == 0:
            result[i] = none_str
        elif value == 1:
            result[i] = nan_str
        elif value == 2:
            result[i] = positive_infty_str
        elif value == 3:
            result[i] = negative_infty_str
        elif not isinstance(value, str):
            raise type_error(value, "computed value", str)
        else:
            # if we get here, we have to deal with actual floats.
            # we split them into int, frac, and exp parts.
            # if the int and exp part exist, we format them as ints.
            int_part: str
            frac_part: str = ""
            exp_part: str = ""
            eidx: int = value.find("e")
            if eidx < 0:
                eidx = value.find("E")
            if eidx >= 0:
                exp_part = exponent_renderer(__toint(int(
                    value[eidx + 1:])).strip()).strip()
                if not isinstance(exp_part, str):
                    raise type_error(exp_part,
                                     "exponent part rendering result", str)
                value = value[:eidx].strip()

            dotidx: int = value.find(".")
            if dotidx <= 0:
                int_part = __toint(int(value))
            else:
                int_part = __toint(int(value[:dotidx]))
                frac_part = value[dotidx:].strip()
            if len(int_part) <= 0:
                int_part = "0"
            result[i] = f"{int_part}{frac_part}{exp_part}"

    return cast(List[str], result)



"""Evaluate the results of the example experiment."""
import os.path as pp
import sys
from statistics import median
from typing import Dict, Final, Optional, Any, List, Set, Union, Callable, \
    Iterable, cast

from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.evaluation.base import TIME_UNIT_FES, TIME_UNIT_MILLIS
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_statistics import EndStatistics
from moptipy.evaluation.tabulate_end_results_impl import \
    tabulate_end_results, command_column_namer
from moptipy.examples.jssp.experiment import EXPERIMENT_INSTANCES
from moptipy.examples.jssp.plots import plot_end_makespans, \
    plot_stat_gantt_charts, plot_progresses, plot_end_makespans_over_param
from moptipy.utils.console import logger
from moptipy.utils.help import help_screen
from moptipy.utils.lang import EN
from moptipy.utils.path import Path
from moptipy.utils.types import type_error

#: The letter mu
LETTER_M: Final[str] = "\u03BC"
#: The letter lambda
LETTER_L: Final[str] = "\u03BB"

#: the pre-defined instance sort keys
__INST_SORT_KEYS: Final[Dict[str, int]] = {
    n: i for i, n in enumerate(EXPERIMENT_INSTANCES)
}


def instance_sort_key(name: str) -> int:
    """
    Get the instance sort key for a given instance name.

    :param name: the instance name
    :returns: the sort key
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    if not name:
        raise ValueError("name must not be empty.")
    if name in __INST_SORT_KEYS:
        return __INST_SORT_KEYS[name]
    return 1000


#: the pre-defined algorithm sort keys
__ALGO_SORT_KEYS: Final[Dict[str, int]] = {
    n: i for i, n in enumerate([
        "1rs", "rs", "hc", "hc_swap2", "hcr_32768_swap2", "hcr", "hcn",
        "hc_swapn", "hcr_65536_swapn", "hcrn", "rls", "rls_swap2",
        "rlsn", "rls_swapn", "rw", "rw_swap2", "rw_swapn"])
}


def algorithm_sort_key(name: str) -> int:
    """
    Get the algorithm sort key for a given algorithm name.

    :param name: the algorithm name
    :returns: the sort key
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    if not name:
        raise ValueError("name must not be empty.")
    if name in __ALGO_SORT_KEYS:
        return __ALGO_SORT_KEYS[name]
    return 1000


#: the algorithm name map
__ALGO_NAME_MAP: Final[Dict[str, str]] = {
    "hc_swap2": "hc", "hcr_32768_swap2": "hcr", "hc_swapn": "hcn",
    "hcr_65536_swapn": "hcrn", "rls_swap2": "rls", "rls_swapn": "rlsn"
}


def algorithm_namer(name: str) -> str:
    """
    Rename algorithm setups.

    :param name: the algorithm's original name
    :returns: the new name of the algorithm
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    if not name:
        raise ValueError("name must not be empty.")
    if name in __ALGO_NAME_MAP:
        return __ALGO_NAME_MAP[name]
    return name


def compute_end_results(results_dir: str,
                        dest_dir: str) -> Path:
    """
    Get the end results, compute them if necessary.

    :param results_dir: the results directory
    :param dest_dir: the destination directory
    :returns: the path to the end results file.
    """
    dest: Final[Path] = Path.directory(dest_dir)
    results_file: Final[Path] = dest.resolve_inside("end_results.txt")
    if results_file.is_file():
        return results_file
    results: Final[List[EndResult]] = []

    source: Final[Path] = Path.directory(results_dir)
    logger(f"loading end results from path '{source}'.")
    EndResult.from_logs(source, results.append)
    if not results:
        raise ValueError(f"Could not find any logs in '{source}'.")
    results.sort()
    logger(f"found {len(results)} log files in path '{source}', storing "
           f"them in file '{results_file}'.")
    rf: Path = EndResult.to_csv(results, results_file)
    if rf != results_file:
        raise ValueError(
            f"results file should be {results_file}, but is {rf}.")
    results_file.enforce_file()
    logger(f"finished writing file '{results_file}'.")
    return results_file


def get_end_results(
        file: str,
        insts: Union[None, Set[str], Callable[[str], bool]] = None,
        algos: Union[None, Set[str], Callable[[str], bool]] = None) \
        -> List[EndResult]:
    """
    Get a specific set of end results..

    :param file: the end results file
    :param insts: only these instances will be included if this parameter is
        provided
    :param algos: only these algorithms will be included if this parameter is
        provided
    """
    def __filter(er: EndResult,
                 ins=None if insts is None else
                 insts.__contains__ if isinstance(insts, Set)
                 else insts,
                 alg=None if algos is None else
                 algos.__contains__ if isinstance(algos, Set)
                 else algos) -> bool:
        if ins is not None:
            if not ins(er.instance):
                return False
        if alg is not None:
            if not alg(er.algorithm):
                return False
        return True

    col: Final[List[EndResult]] = []
    EndResult.from_csv(file=file, consumer=col.append, filterer=__filter)
    if len(col) <= 0:
        raise ValueError(
            f"no end results for instances {insts} and algorithms {algos}.")
    return col


def compute_end_statistics(end_results_file: str,
                           dest_dir: str) -> Path:
    """
    Get the end result statistics, compute them if necessary.

    :param end_results_file: the end results file
    :param dest_dir: the destination directory
    :returns: the path to the end result statistics file.
    """
    dest: Final[Path] = Path.directory(dest_dir)
    stats_file: Final[Path] = dest.resolve_inside("end_statistics.txt")
    if stats_file.is_file():
        return stats_file

    results: Final[List[EndResult]] = get_end_results(end_results_file)
    if len(results) <= 0:
        raise ValueError("end results cannot be empty")
    stats: Final[List[EndStatistics]] = []
    EndStatistics.from_end_results(results, stats.append)
    if len(stats) <= 0:
        raise ValueError("end result statistics cannot be empty")
    stats.sort()

    sf: Path = EndStatistics.to_csv(stats, stats_file)
    if sf != stats_file:
        raise ValueError(f"stats file should be {stats_file} but is {sf}")
    stats_file.enforce_file()
    logger(f"finished writing file '{stats_file}'.")
    return stats_file


def table(end_results: Path, algos: List[str], dest: Path) -> None:
    """
    Tabulate the end results.

    :param end_results: the path to the end results
    :param algos: the algorithms
    :param dest: the directory
    """
    EN.set_current()
    n: Final[str] = algorithm_namer(algos[0])
    tabulate_end_results(
        end_results=get_end_results(end_results, algos=set(algos)),
        file_name=f"end_results_{n}", dir_name=dest,
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        col_namer=command_column_namer,
        algorithm_namer=algorithm_namer,
        use_lang=False)


def makespans(end_results: Path, algos: List[str], dest: Path) -> None:
    """
    Plot the end makespans.

    :param end_results: the path to the end results
    :param algos: the algorithms
    :param dest: the directory
    """
    n: Final[str] = algorithm_namer(algos[0])
    plot_end_makespans(
        end_results=get_end_results(end_results, algos=set(algos)),
        name_base=f"makespan_scaled_{n}", dest_dir=dest,
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        algorithm_namer=algorithm_namer)


def gantt(end_results: Path, algo: str, dest: Path, source: Path,
          best: bool = False,
          insts: Optional[Iterable[str]] = None) -> None:
    """
    Plot the median Gantt charts.

    :param end_results: the path to the end results
    :param algo: the algorithm
    :param dest: the directory
    :param source: the source directory
    :param best: should we plot the best instance only (or, otherwise, the
        median)
    :param insts: the instances to use
    """
    n: Final[str] = algorithm_namer(algo)
    plot_stat_gantt_charts(
        get_end_results(end_results, algos={algo},
                        insts=None if insts is None else set(insts)),
        name_base=f"best_gantt_{n}" if best else f"gantt_{n}",
        dest_dir=dest,
        results_dir=source,
        instance_sort_key=instance_sort_key,
        statistic=cast(
            Callable[[Iterable[Union[int, float]]], Union[int, float]],
            min if best else median))


def progress(algos: List[str], dest: Path, source: Path,
             log: bool = True, millis: bool = True) -> None:
    """
    Plot the median Gantt charts.

    :param algos: the algorithms
    :param dest: the directory
    :param source: the source directory
    :param log: is the time logarithmically scaled?
    :param millis: is the time measured in milliseconds?
    """
    n: str = f"progress_{algorithm_namer(algos[0])}_"
    if log:
        n = n + "log_"
    if millis:
        unit = TIME_UNIT_MILLIS
        n = n + "T"
    else:
        unit = TIME_UNIT_FES
        n = n + "FEs"
    plot_progresses(results_dir=source,
                    algorithms=algos,
                    name_base=n,
                    dest_dir=dest,
                    log_time=log,
                    time_unit=unit,
                    algorithm_namer=algorithm_namer,
                    instance_sort_key=instance_sort_key,
                    algorithm_sort_key=algorithm_sort_key)


def makespans_over_param(
        end_results: Path,
        selector: Callable[[str], bool],
        x_getter: Callable[[EndStatistics], Union[int, float]],
        name_base: str,
        dest_dir: str,
        x_axis: Union[AxisRanger, Callable[[], AxisRanger]]
        = AxisRanger,
        x_label: Optional[str] = None,
        algo_getter: Optional[Callable[[str], str]] = None,
        title: Optional[str] = None) -> List[Path]:
    """
    Plot the performance over a parameter.

    :param end_results: the end results path
    :param selector: the selector for algorithms
    :param name_base: the basic name
    :param dest_dir: the destination directory
    :param x_getter: the function computing the x-value for each statistics
        object
    :param x_axis: the axis ranger
    :param x_label: the x-axis label
    :param algo_getter: the optional algorithm name getter (use `name_base` if
        unspecified)
    :param title: the optional title (use `name_base` if unspecified)
    :returns: the list of generated files
    """
    def _algo_name_getter(es: EndStatistics,
                          n=name_base, g=algo_getter) -> str:
        return n if g is None else g(es.algorithm)

    return plot_end_makespans_over_param(
        end_results=get_end_results(end_results, algos=selector),
        x_getter=x_getter, name_base=f"{name_base}_results",
        dest_dir=dest_dir, title=name_base if title is None else title,
        algorithm_getter=_algo_name_getter,  # type: ignore
        instance_sort_key=instance_sort_key,
        algorithm_sort_key=algorithm_sort_key,
        x_axis=x_axis, x_label=x_label,
        plot_single_instances=algo_getter is None)


def evaluate_experiment(results_dir: str = pp.join(".", "results"),
                        dest_dir: Optional[str] = None) -> None:
    """
    Evaluate the experiment.

    :param results_dir: the results directory
    :param dest_dir: the destination directory
    """
    source: Final[Path] = Path.directory(results_dir)
    dest: Final[Path] = Path.path(dest_dir if dest_dir else
                                  pp.join(source, "..", "evaluation"))
    dest.ensure_dir_exists()
    logger(f"Beginning evaluation from '{source}' to '{dest}'.")

    end_results: Final[Path] = compute_end_results(source, dest)
    if not end_results:
        raise ValueError("End results path is empty??")
    end_stats: Final[Path] = compute_end_statistics(end_results, dest)
    if not end_stats:
        raise ValueError("End stats path is empty??")

    logger("Now evaluating the single random sampling algorithm `1rs`.")
    table(end_results, ["1rs"], dest)
    makespans(end_results, ["1rs"], dest)
    gantt(end_results, "1rs", dest, source)

    logger("Now evaluating the multi-random sampling algorithm `rs`.")
    table(end_results, ["rs", "1rs"], dest)
    makespans(end_results, ["rs", "1rs"], dest)
    gantt(end_results, "rs", dest, source)
    progress(["rs"], dest, source)
    progress(["rs"], dest, source, log=False)

    logger("Now evaluating the hill climbing algorithm `hc`.")
    table(end_results, ["hc_swap2", "rs"], dest)
    makespans(end_results, ["hc_swap2", "rs"], dest)
    gantt(end_results, "hc_swap2", dest, source)
    progress(["hc_swap2", "rs"], dest, source)
    progress(["hc_swap2", "rs"], dest, source, millis=False)

    logger("Now evaluating the hill climbing algorithm with "
           "restarts 'hcr' on 'swap2'.")
    makespans_over_param(
        end_results,
        lambda an: an.startswith("hcr_") and an.endswith("_swap2"),
        lambda es: int(es.algorithm.split("_")[1]),
        "hcr_L_swap2", dest,
        lambda: AxisRanger(log_scale=True, log_base=2.0), "L")
    table(end_results, ["hcr_32768_swap2", "hc_swap2", "rs"], dest)
    makespans(end_results, ["hcr_32768_swap2", "hc_swap2", "rs"], dest)
    gantt(end_results, "hcr_32768_swap2", dest, source)
    progress(["hcr_32768_swap2", "hc_swap2", "rs"], dest, source)

    logger("Now evaluating the hill climbing algorithm with 'swapn'.")
    table(end_results, ["hc_swapn", "hcr_32768_swap2", "hc_swap2"], dest)
    makespans(end_results, ["hc_swapn", "hcr_32768_swap2", "hc_swap2"], dest)
    progress(["hc_swapn", "hcr_32768_swap2", "hc_swap2"], dest, source)
    progress(["hc_swapn", "hcr_32768_swap2", "hc_swap2"], dest, source,
             millis=False)

    logger("Now evaluating the hill climbing algorithm with "
           "restarts 'hcr' on 'swapn'.")
    makespans_over_param(
        end_results,
        lambda an: an.startswith("hcr_") and an.endswith("_swapn"),
        lambda es: int(es.algorithm.split("_")[1]),
        "hcr_L_swapn", dest,
        lambda: AxisRanger(log_scale=True, log_base=2.0), "L")
    table(end_results, ["hcr_65536_swapn", "hc_swapn",
                        "hcr_32768_swap2"], dest)
    makespans(end_results, ["hcr_65536_swapn", "hc_swapn",
                            "hcr_32768_swap2"], dest)
    progress(["hcr_65536_swapn", "hc_swapn",
              "hcr_32768_swap2"], dest, source)

    logger("Now evaluating the RLS algorithm with 'swap2' and 'swapn'.")
    table(end_results, ["rls_swapn", "rls_swap2",
                        "hcr_32768_swap2", "hcr_65536_swapn"], dest)
    makespans(end_results, ["rls_swapn", "rls_swap2", "hcr_32768_swap2",
                            "hcr_65536_swapn"], dest)
    gantt(end_results, "rls_swap2", dest, source)
    progress(["rls_swapn", "rls_swap2", "hcr_32768_swap2",
              "hcr_65536_swapn"], dest, source)
    gantt(end_results, "rls_swap2", dest, source, True, ["ta70"])

    logger(f"Finished evaluation from '{source}' to '{dest}'.")


# Evaluate experiment if run as script
if __name__ == '__main__':
    help_screen(
        "JSSP Experiment Evaluator", __file__,
        "Evaluate the results of the JSSP experiment.",
        [("results_dir",
          "the directory with the results of the JSSP experiment",
          True),
         ("dest_dir",
          "the directory where the evaluation results should be stored",
          True)])
    mkwargs: Dict[str, Any] = {}
    if len(sys.argv) > 1:
        results_dir_str: Final[str] = sys.argv[1]
        mkwargs["results_dir"] = results_dir_str
        logger(f"Set results_dir '{results_dir_str}'.")
        if len(sys.argv) > 2:
            dest_dir_str: Final[str] = sys.argv[2]
            mkwargs["dest_dir"] = dest_dir_str
            logger(f"Set dest_dir to '{dest_dir_str}'.")

    evaluate_experiment(**mkwargs)  # invoke the experiment evaluation



"""
The implementation of a (mu+lambda) Evolutionary Algorithm.

This is the basic `mu+lambda`-EA works as follows:

1. Start with a list `lst` of `mu` random records and `lambda` blank records.
2. In each iteration:
    2.1. Use the `mu` first records as input to the search operators to
         generate `lambda` new points in the search space:
         For each new point to be created, the binary operator is applied
         with probability `0<=br<=1` and the unary operator is used otherwise.

         Each of the `mu` records has the same chance to produce such a new
         point in the search space, but no record can be used again (in one
         iteration) until all other `mu-1` selected records have produced at
         least one point. In other words, if `lambda > mu`, then each of the
         `mu` selected record will produce at least `lambda // mu` offspring
         and at most `1 + lambda // mu`.
    2.1. Sort the list `lst` according to the objective value of the record.
         Ties are broken such that younger record are preferred over older
         ones (if they have the same objective value).

If `mu=1`, `lambda=1`, and `br=0`, then this algorithm is exactly equivalent
to the :class:`~moptipy.algorithms.so.rls.RLS` if the same unary and nullary
operator are used. It is only a bit slower due to the additional overhead of
maintaining a list of records.
"""
from math import isfinite
from typing import Final, Union, Callable, List, cast, Optional

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm2
from moptipy.api.operators import Op0, Op1, Op2
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import num_to_str_for_name
from moptipy.utils.types import type_error


# start record
class _Record:
    """
    A point in the search space, its quality, and creation time.

    A record stores a point in the search space :attr:`x` together with
    the corresponding objective value :attr:`f`. It also stores a
    "iteration index" :attr:`it`, i.e., the time when the point was
    created or modified.

    This allows for representing and storing solutions in a population.
    If the population is sorted, then records with better objective
    value will be moved to the beginning of the list. Ties are broken
    such that younger individuals (with higher :attr:`it` value) are
    preferred.
    """

    def __init__(self, x, f: Union[int, float]):
        """
        Create the record.

        :param x: the data structure for a point in the search space
        :param f: the corresponding objective value
        """
        #: the point in the search space
        self.x: Final = x
        #: the objective value corresponding to x
        self.f: Union[int, float] = f
        #: the iteration index when the record was created
        self.it: int = 0

    def __lt__(self, other) -> bool:
        """
        Precedence if 1) better or b) equally good but younger.

        :param other: the other record
        :returns: `True` if this record has a better objective value
            (:attr:`f`) or if it has the same objective value but is newer,
            i.e., has a larger :attr:`it` value

        >>> r1 = _Record(None, 10)
        >>> r2 = _Record(None, 9)
        >>> r1 < r2
        False
        >>> r2 < r1
        True
        >>> r1.it = 22
        >>> r2.f = r1.f
        >>> r2.it = r1.it
        >>> r1 < r2
        False
        >>> r2 < r1
        False
        >>> r2.it = r1.it + 1
        >>> r1 < r2
        False
        >>> r2 < r1
        True
        """
        f1: Final[Union[int, float]] = self.f
        f2: Final[Union[int, float]] = other.f
        return (f1 < f2) or ((f1 == f2) and (self.it > other.it))
# end record


def _int_0(_: int) -> int:
    """
    Return an integer with value `0`.

    :retval 0: always
    """
    return 0


def _float_0() -> float:
    """
    Return a float with value `0.0`.

    :retval 0.0: always
    """
    return 0.0


# start nobinary
class EA(Algorithm2):
    """
    The EA is a population-based algorithm using unary and binary operators.

    It starts with a list of `mu` randomly initialized solutions. In each
    step, it retains the `mu` best solutions and generates `lambda` new
    solutions from them using the unary or binary search operator. From the
    joint set of `mu+lambda` solutions, it again selects the best `mu` ones
    for the next iteration. And so on.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the EA to an optimization problem.

        :param process: the black-box process object
        """
        mu: Final[int] = self.__mu  # mu: number of best solutions kept
        lst_size: Final[int] = mu + self.__lambda  # size = mu + lambda
        # Omitted for brevity: store function references in variables
        # end nobinary
        random: Final[Generator] = process.get_random()  # random gen
        create: Final[Callable] = process.create  # create x container
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator
        op2: Final[Callable] = self.op2.op2  # the binary operator
        br: Final[float] = self.__br  # the rate at which to use op2
        should_terminate: Final[Callable] = process.should_terminate
        r0i: Final[Callable[[int], int]] = cast(   # only if m > 1, we
            Callable[[int], int], random.integers  # need random
            if mu > 1 else _int_0)                 # indices
        r01: Final[Callable[[], float]] = cast(  # only if 0<br<1, we
            Callable[[], float],                 # need random floats
            random.random if 0 < br < 1 else _float_0)
        # start nobinary
        # create list of mu random records and lambda empty records
        lst: Final[List] = [None] * lst_size  # pre-allocate list
        f: Union[int, float] = 0  # variable to hold objective values
        for i in range(lst_size):  # fill list of size mu+lambda
            x = create()  # by creating point in search space
            if i < mu:  # only the first mu records are initialized by
                op0(random, x)  # applying nullary operator = randomize
                if should_terminate():  # should we quit?
                    return   # computational budget exhausted -> quit
                f = evaluate(x)  # continue? ok, evaluate new solution
            lst[i] = _Record(x, f)  # create and store record

        iteration: int = 1  # The first "real" iteration has index 1
        while True:  # lst: keep 0..mu-1, overwrite mu..mu+lambda-1
            end: int = mu  # the end index for available old records
            for oi in range(mu, lst_size):  # for all lambda offspring
                if should_terminate():  # only continue if we still...
                    return  # have sufficient budget ... otherwise quit

                offspring: _Record = lst[oi]  # pick offspring
                x = offspring.x  # the point in search space we work on
                offspring.it = iteration  # mark as member of new iter.

                pi: int = r0i(end)  # random record in 0..end-1=unused
                rec: _Record = lst[pi]  # pick the unused record
                if end > 1:  # swap rec to the end to not use again
                    end = end - 1  # reduce number of unused records
                    lst[end], lst[pi] = rec, lst[end]  # rec <-> end
                else:  # oh: we have used all records
                    end = mu  # then lambda >= mu and we re-use them
                # end nobinary
                # start binary
                if r01() < br:  # apply binary operator at rate br
                    pi = r0i(end)  # random record in 0..end-1
                    rec2: _Record = lst[pi]  # get second unused record
                    if end > 1:  # swap rec2 to end to not use again
                        end = end - 1  # reduce num. of unused records
                        lst[end], lst[pi] = rec2, lst[end]
                    else:  # oh: we have used all records?
                        end = mu  # then lambda>=mu and we re-use them
                    op2(random, x, rec.x, rec2.x)  # apply binary op
                    offspring.f = evaluate(x)  # evaluate new point
                    continue  # below is "else" part with unary operat.
                # end binary
                # start nobinary
                op1(random, x, rec.x)  # apply unary operator to rec.x
                offspring.f = evaluate(x)  # evaluate new point

            iteration = iteration + 1  # step iteration counter
            lst.sort()  # sort list: best records come first
# end nobinary

    def __init__(self, op0: Op0,
                 op1: Optional[Op1] = None,
                 op2: Optional[Op2] = None,
                 mu: int = 1, lambda_: int = 1,
                 br: Optional[float] = None) -> None:
        """
        Create the Evolutionary Algorithm (EA) without binary crossover.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param op2: the binary search operator
        :param mu: the number of best solutions to survive in each generation
        :param lambda_: the number of offspring in each generation
        :param br: the rate at which the binary operator is applied
        """
        if op1 is None:
            op1 = Op1()
            if br is None:
                br = 1.0
            elif br != 1.0:
                raise ValueError(
                    f"if op1==None, br must be None or 1.0, but is {br}.")
        elif op1.__class__ == Op1:
            if br is None:
                br = 1.0
            elif br != 1.0:
                raise ValueError(
                    f"if op1 is Op1, br must be None or 1.0, but is {br}.")
        elif (br is not None) and (br == 1.0):
            op1 = Op1()

        if op2 is None:
            op2 = Op2()
            if br is None:
                br = 0.0
            elif br != 0.0:
                raise ValueError(
                    f"if op2==None, br must be None or 0.0, but is {br}.")
        elif op2.__class__ == Op2:
            if br is None:
                br = 0.0
            elif br != 0.0:
                raise ValueError(
                    f"if op2 is Op2, br must be None or 0.0, but is {br}.")
        elif (br is not None) and (br == 0.0):
            op2 = Op2()
        elif mu == 1:
            if br is None:
                br = 0.0
                op2 = Op2()

        if br is None:
            br = 0.2

        super().__init__(
            f"ea_{mu}_{lambda_}_{num_to_str_for_name(br)}"
            if 0 < br < 1 else f"ea_{mu}_{lambda_}", op0, op1, op2)

        if not isinstance(mu, int):
            raise type_error(mu, "mu", int)
        if not (0 < mu <= 1_000_000):
            raise ValueError(f"invalid mu={mu}, must be in 1..1_000_000.")
        if not isinstance(lambda_, int):
            raise type_error(lambda_, "lambda", int)
        if not (0 < lambda_ <= 1_000_000):
            raise ValueError(
                f"invalid lambda={lambda_}, must be in 1..1_000_000.")
        if not isinstance(br, float):
            raise type_error(br, "br", float)
        if not (isfinite(br) and (0.0 <= br <= 1.0)):
            raise ValueError(f"invalid br={br}, must be in [0, 1].")
        if (br > 0.0) and (mu <= 1):
            raise ValueError(
                f"if br (={br}) > 0, then mu (={mu}) must be > 1.")
        #: the number of records to survive in each generation
        self.__mu: Final[int] = mu
        #: the number of offsprings per generation
        self.__lambda: Final[int] = lambda_
        #: the rate at which the binary operator is applied
        self.__br: Final[float] = br

    def log_parameters_to(self, logger: KeyValueLogSection):
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("mu", self.__mu)
        logger.key_value("lambda", self.__lambda)
        logger.key_value("br", self.__br)



"""Different ways to transform and slice processes."""
from typing import Final, Callable, Union, Any, TypeVar, Optional

import numpy as np

from moptipy.api.mo_process import MOProcess
from moptipy.api.process import Process, check_max_fes
from moptipy.utils.types import type_error

#: the type variable for single- and multi-objective processes.
T = TypeVar('T', Process, MOProcess)


class __ForFEs(Process):
    """A process searching for a fixed amount of FEs."""

    def __init__(self, owner: Process, max_fes: int):
        super().__init__()
        if not isinstance(owner, Process):
            raise type_error(owner, "owner", Process)
        #: the owning process
        self._owner: Final[Process] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.has_best = owner.has_best  # type: ignore
        self.get_copy_of_best_x = owner.get_copy_of_best_x  # type: ignore
        self.get_best_f = owner.get_best_f  # type: ignore
        #: the maximum FEs
        self.max_fes: Final[int] = check_max_fes(max_fes)
        #: the FEs that we still have left
        self.__fes_left: int = max_fes
        #: did we terminate?
        self.__terminated: bool = False
        #: the fast call to the owner's should_terminate method
        self.__should_terminate: Final[Callable[[], bool]] \
            = owner.should_terminate
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], Union[int, float]]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, Union[int, float]], None]] \
            = owner.register
        #: the start fe
        self.__start_fe: Final[int] = owner.get_consumed_fes()

    def should_terminate(self) -> bool:
        return self.__terminated or self.__should_terminate()

    def terminate(self) -> None:
        self.__terminated = True

    def evaluate(self, x) -> Union[float, int]:
        f: Final[Union[int, float]] = self.__evaluate(x)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True
        return f

    def register(self, x, f: Union[int, float]) -> None:
        self.__register(x, f)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True

    def get_consumed_fes(self) -> int:
        return self.max_fes - self.__fes_left

    def get_last_improvement_fe(self) -> int:
        return max(1 if self.__fes_left < self.max_fes else 0,
                   self._owner.get_last_improvement_fe() - self.__start_fe)

    def get_max_fes(self) -> int:
        return self.max_fes

    def __str__(self) -> str:
        return f"forFEs_{self.max_fes}_{self._owner}"


class __ForFEsMO(MOProcess):
    """A process searching for a fixed amount of FEs."""

    def __init__(self, owner: MOProcess, max_fes: int):
        super().__init__()
        if not isinstance(owner, MOProcess):
            raise type_error(owner, "owner", MOProcess)
        #: the owning process
        self._owner: Final[MOProcess] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.has_best = owner.has_best  # type: ignore
        self.get_copy_of_best_x = owner.get_copy_of_best_x  # type: ignore
        self.get_best_f = owner.get_best_f  # type: ignore
        self.get_archive = owner.get_archive  # type: ignore
        self.check_in = owner.check_in  # type: ignore
        self.f_create = owner.f_create  # type: ignore
        self.f_validate = owner.f_validate  # type: ignore
        self.f_dtype = owner.f_dtype  # type: ignore
        self.f_dominates = owner.f_dominates  # type: ignore
        self.f_dimension = owner.f_dimension  # type: ignore
        #: the maximum FEs
        self.max_fes: Final[int] = check_max_fes(max_fes)
        #: the FEs that we still have left
        self.__fes_left: int = max_fes
        #: did we terminate?
        self.__terminated: bool = False
        #: the fast call to the owner's should_terminate method
        self.__should_terminate: Final[Callable[[], bool]] \
            = owner.should_terminate
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], Union[int, float]]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, Union[int, float]], None]] \
            = owner.register
        #: the evaluation wrapper
        self.__f_evaluate: Final[Callable[
            [Any, np.ndarray], Union[int, float]]] = owner.f_evaluate
        #: the start fe
        self.__start_fe: Final[int] = owner.get_consumed_fes()

    def should_terminate(self) -> bool:
        return self.__terminated or self.__should_terminate()

    def terminate(self) -> None:
        self.__terminated = True

    def evaluate(self, x) -> Union[float, int]:
        f: Final[Union[int, float]] = self.__evaluate(x)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True
        return f

    def register(self, x, f: Union[int, float]) -> None:
        self.__register(x, f)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True

    def f_evaluate(self, x, fs: np.ndarray) -> Union[float, int]:
        f: Final[Union[int, float]] = self.__f_evaluate(x, fs)
        fel: Final[int] = self.__fes_left - 1
        self.__fes_left = fel
        if fel <= 0:
            self.__terminated = True
        return f

    def get_consumed_fes(self) -> int:
        return self.max_fes - self.__fes_left

    def get_last_improvement_fe(self) -> int:
        return max(1 if self.__fes_left < self.max_fes else 0,
                   self._owner.get_last_improvement_fe() - self.__start_fe)

    def get_max_fes(self) -> int:
        return self.max_fes

    def __str__(self) -> str:
        return f"forFEsMO_{self.max_fes}_{self._owner}"


def for_fes(process: T, max_fes: int) -> T:
    """
    Create a sub-process that can run for the given number of FEs.

    :param process: the original process
    :param max_fes: the maximum number of objective function evaluations
    :returns: the sub-process that will terminate after `max_fes` FEs and that
        forwards all other calls the `process`.
    """
    if not isinstance(max_fes, int):
        raise type_error(max_fes, "max_fes", int)
    if max_fes <= 0:
        raise ValueError(f"max_fes cannot be {max_fes}!")
    return __ForFEsMO(process, max_fes) if isinstance(process, MOProcess) \
        else __ForFEs(process, max_fes)


class __FromStatingPoint(Process):
    """A process searching from a given point."""

    def __init__(self, owner: Process, in_and_out_x: Any,
                 f: Union[int, float]):
        """
        Create a sub-process searching from one starting point.

        :param owner: the owning process
        :param in_and_out_x: the input solution record, which will be
            overwritten with the best encountered solution
        :param f: the objective value corresponding to `in_and_out`
        """
        super().__init__()
        if not isinstance(owner, Process):
            raise type_error(owner, "owner", Process)
        #: the owning process
        self._owner: Final[Process] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.should_terminate = owner.should_terminate  # type: ignore
        self.terminate = owner.terminate  # type: ignore
        #: the best solution
        self.__best_x: Final[Any] = in_and_out_x
        #: the best-so-far solution
        self.__best_f: Union[int, float] = f
        #: the last improvement fe
        self.__last_improvement_fe: int = 0
        #: the consumed FEs
        self.__fes: int = 0
        mfes: Optional[int] = owner.get_max_fes()
        if mfes is not None:
            mfes = mfes - owner.get_consumed_fes()
        #: the maximum permitted FEs
        self.__max_fes: Final[Optional[int]] = mfes
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], Union[int, float]]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, Union[int, float]], None]] \
            = owner.register

    def has_best(self) -> bool:
        return True

    def get_copy_of_best_x(self, x) -> None:
        self.copy(x, self.__best_x)

    def get_best_f(self) -> Union[int, float]:
        return self.__best_f

    def evaluate(self, x) -> Union[float, int]:
        self.__fes = fe = self.__fes + 1
        f: Final[Union[int, float]] = self.__evaluate(x)
        if f < self.__best_f:
            self.__best_f = f
            self.copy(self.__best_x, x)
            self.__last_improvement_fe = fe
        return f

    def register(self, x, f: Union[int, float]) -> None:
        self.__register(x, f)
        self.__fes = fe = self.__fes + 1
        if f < self.__best_f:
            self.__best_f = f
            self.copy(self.__best_x, x)
            self.__last_improvement_fe = fe

    def get_consumed_fes(self) -> int:
        return max(1, self.__fes)

    def get_last_improvement_fe(self) -> int:
        return self.__last_improvement_fe

    def get_max_fes(self) -> Optional[int]:
        return self.__max_fes

    def __str__(self) -> str:
        return f"fromStart_{self._owner}"


def from_starting_point(owner: Process, in_and_out_x: Any,
                        f: Union[int, float]) -> Process:
    """
    Create a sub-process searching from one starting point.

    :param owner: the owning process
    :param in_and_out_x: the input solution record, which will be
        overwritten with the best encountered solution
    :param f: the objective value corresponding to `in_and_out`
    """
    return __FromStatingPoint(owner, in_and_out_x, f)


class _InternalTerminationError(Exception):
    """A protected internal termination error."""


class __WithoutShouldTerminate(Process):
    """A process allowing algorithm execution ignoring `should_terminate`."""

    def __init__(self, owner: Process):
        super().__init__()
        if not isinstance(owner, Process):
            raise type_error(owner, "owner", Process)
        #: the owning process
        self._owner: Final[Process] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        a = owner.get_last_improvement_fe  # type: ignore
        self.get_last_improvement_fe = a  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.has_best = owner.has_best  # type: ignore
        self.get_copy_of_best_x = owner.get_copy_of_best_x  # type: ignore
        self.get_best_f = owner.get_best_f  # type: ignore
        self.should_terminate = owner.should_terminate  # type: ignore
        self.terminate = owner.terminate  # type: ignore
        self.get_max_fes = owner.get_max_fes  # type: ignore
        self.get_consumed_fes = owner.get_consumed_fes  # type: ignore
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], Union[int, float]]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, Union[int, float]], None]] \
            = owner.register

    def evaluate(self, x) -> Union[float, int]:
        if self.should_terminate():
            raise _InternalTerminationError()
        return self.__evaluate(x)

    def register(self, x, f: Union[int, float]) -> None:
        if self.should_terminate():
            raise _InternalTerminationError()
        self.__register(x, f)

    def __str__(self) -> str:
        return f"protect_{self._owner}"

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        return True


class __WithoutShouldTerminateMO(MOProcess):
    """A process allowing algorithm execution ignoring `should_terminate`."""

    def __init__(self, owner: Process):
        super().__init__()
        if not isinstance(owner, MOProcess):
            raise type_error(owner, "owner", MOProcess)
        #: the owning process
        self._owner: Final[MOProcess] = owner
        self.get_random = owner.get_random  # type: ignore
        a = owner.get_consumed_time_millis  # type: ignore
        self.get_consumed_time_millis = a  # type: ignore
        a = owner.get_max_time_millis  # type: ignore
        self.get_max_time_millis = a  # type: ignore
        a = owner.get_last_improvement_time_millis  # type: ignore
        self.get_last_improvement_time_millis = a  # type: ignore
        a = owner.get_last_improvement_fe  # type: ignore
        self.get_last_improvement_fe = a  # type: ignore
        self.add_log_section = owner.add_log_section  # type: ignore
        self.lower_bound = owner.lower_bound  # type: ignore
        self.upper_bound = owner.upper_bound  # type: ignore
        self.create = owner.create  # type: ignore
        self.copy = owner.copy  # type: ignore
        self.to_str = owner.to_str  # type: ignore
        self.is_equal = owner.is_equal  # type: ignore
        self.from_str = owner.from_str  # type: ignore
        self.validate = owner.validate  # type: ignore
        self.n_points = owner.n_points  # type: ignore
        self.has_best = owner.has_best  # type: ignore
        self.get_copy_of_best_x = owner.get_copy_of_best_x  # type: ignore
        self.get_best_f = owner.get_best_f  # type: ignore
        self.should_terminate = owner.should_terminate  # type: ignore
        self.terminate = owner.terminate  # type: ignore
        self.get_max_fes = owner.get_max_fes  # type: ignore
        self.get_consumed_fes = owner.get_consumed_fes  # type: ignore
        self.f_create = owner.f_create  # type: ignore
        self.f_validate = owner.f_validate  # type: ignore
        self.f_dtype = owner.f_dtype  # type: ignore
        self.f_dominates = owner.f_dominates  # type: ignore
        self.f_dimension = owner.f_dimension  # type: ignore
        self.get_archive = owner.get_archive  # type: ignore
        self.check_in = owner.check_in  # type: ignore
        #: the fast call to the owner's evaluate method
        self.__evaluate: Final[Callable[[Any], Union[int, float]]] \
            = owner.evaluate
        #: the fast call to the owner's register method
        self.__register: Final[Callable[[Any, Union[int, float]], None]] \
            = owner.register
        #: the fast call to the owner's f_evaluate method
        self.__f_evaluate: Final[Callable[
            [Any, np.ndarray], Union[int, float]]] \
            = owner.f_evaluate

    def evaluate(self, x) -> Union[float, int]:
        if self.should_terminate():
            raise _InternalTerminationError()
        return self.__evaluate(x)

    def register(self, x, f: Union[int, float]) -> None:
        if self.should_terminate():
            raise _InternalTerminationError()
        self.__register(x, f)

    def f_evaluate(self, x, fs: np.ndarray) -> Union[int, float]:
        if self.should_terminate():
            raise _InternalTerminationError()
        return self.__f_evaluate(x, fs)

    def __str__(self) -> str:
        return f"protectMO_{self._owner}"

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        return True


def without_should_terminate(algorithm: Callable[[T], Any], process: T):
    """
    Apply an algorithm that does not call `should_terminate` to a process.

    If we use an algorithm from an external library, this algorithm may ignore
    the proper usage of our API. With this method, we try to find a way to
    make sure that these calls are consistent with the termination criterion
    of the `moptipy` API.

    Before calling :meth:`~moptipy.api.process.Process.evaluate`,
    :meth:`~moptipy.api.process.Process.register`, or
    :meth:`~moptipy.api.mo_problem.MOProblem.f_evaluate`, an optimization
    algorithm must check if it should instead stop via
    :meth:`~moptipy.api.process.Process.should_terminate`. If the process
    called :meth:`~moptipy.api.process.Process.should_terminate` and was
    told to stop but did invoke the evaluation routines anyway, an
    exception will be thrown and the process force-terminates. If the
    process did not call
    :meth:`~moptipy.api.process.Process.should_terminate` but was
    supposed to stop, the results of
    :meth:`~moptipy.api.process.Process.evaluate` may be arbitrary (or
    positive infinity).

    This function here can be used to deal with processes that do not invoke
    :meth:`~moptipy.api.process.Process.should_terminate`. It will invoke
    this method by itself before
    :meth:`~moptipy.api.process.Process.evaluate`,
    :meth:`~moptipy.api.process.Process.register`, and
    :meth:`~moptipy.api.mo_problem.MOProblem.f_evaluate` and terminate the
    algorithm with an exception if necessary. It will then catch the
    exception and bury it.

    Thus, we can now use algorithms that ignore our termination criteria and
    still force them to terminate when they should.

    :param algorithm: the algorithm
    :param process: the optimization process
    """
    try:
        with __WithoutShouldTerminateMO(process) \
                if isinstance(process, MOProcess) \
                else __WithoutShouldTerminate(process) as proc:
            algorithm(proc)
    except _InternalTerminationError:
        pass  # the internal error is ignored



"""Infrastructure to created structured text, like Markdown and LaTeX."""

from io import TextIOBase
from typing import Final

from moptipy.utils.path import Path
from moptipy.utils.types import type_error
from moptipy.utils.lang import Lang


class FormattedStr(str):
    """A subclass of `str` capable of holding formatting information."""

    #: should this string be formatted in bold face?
    bold: bool
    #: should this string be formatted in italic face?
    italic: bool
    #: should this string be formatted in code face?
    code: bool

    def __new__(cls, value, bold: bool = False, italic: bool = False,
                code: bool = False):
        """
        Construct the object.

        :param value: the string value
        :param bold: should the format be bold face?
        :param italic: should the format be italic face?
        :param code: should the format be code face?
        """
        if not isinstance(bold, bool):
            raise type_error(bold, "bold", bool)
        if not isinstance(italic, bool):
            raise type_error(italic, "italic", bool)
        if not isinstance(code, bool):
            raise type_error(code, "code", bool)
        if bold or italic or code:
            ret = super(FormattedStr, cls).__new__(cls, value)
            ret.bold = bold
            ret.italic = italic
            ret.code = code
            return ret
        return value

    @staticmethod
    def add_format(s: str, bold: bool = False, italic: bool = False,
                   code: bool = False) -> str:
        """
        Add the given format to the specified string.

        :param s: the string
        :param bold: should the format be bold face?
        :param italic: should the format be italic face?
        :param code: should the format be code face?

        >>> from typing import cast
        >>> st = "abc"
        >>> type(st)
        <class 'str'>
        >>> fs = cast(FormattedStr, FormattedStr.add_format(st, bold=True))
        >>> type(fs)
        <class 'moptipy.utils.text_format.FormattedStr'>
        >>> fs.bold
        True
        >>> fs.italic
        False
        >>> fs = cast(FormattedStr, FormattedStr.add_format(fs, italic=True))
        >>> fs.bold
        True
        >>> fs.italic
        True
        """
        if isinstance(s, FormattedStr):
            bold = bold or s.bold
            italic = italic or s.italic
            code = code or s.code
            if (bold != s.bold) or (italic != s.italic) or (code != s.code):
                return FormattedStr(s, bold, italic, code)
            return s
        if not isinstance(s, str):
            raise type_error(s, "s", str)
        if bold or italic or code:
            return FormattedStr(s, bold, italic, code)
        return s


class TextFormatDriver:
    """
    A base class for text format drivers.

    Table drivers allow us to render the structured text. It used, for
    example, by instances of :class:`~moptipy.utils.table.Table` to write
    tabular data to a text format stream.
    """

    def begin_table_body(self, stream: TextIOBase, cols: str) -> None:
        """
        Write the beginning of the table body.

        :param stream: the stream to write to
        :param cols: the column definition
        """

    def end_table_body(self, stream: TextIOBase, cols: str) -> None:
        """
        Write the ending of the table body.

        :param stream: the stream to write to
        :param cols: the column definition
        """

    def begin_table_header(self, stream: TextIOBase, cols: str) -> None:
        """
        Begin the header row of the table.

        :param stream: the stream to write to
        :param cols: the column definition
        """

    def end_table_header(self, stream: TextIOBase, cols: str) -> None:
        """
        End the header row of the table.

        :param stream: the stream to write to
        :param cols: the column definition
        """

    def begin_table_section(self, stream: TextIOBase, cols: str,
                            section_index: int) -> None:
        """
        Begin a new section of the table.

        :param stream: the stream to write to
        :param cols: the column definition
        :param section_index: the index of the section, `0` for the first
            section
        """

    def end_table_section(self, stream: TextIOBase, cols: str,
                          section_index: int,
                          n_rows: int) -> None:
        """
        End a section of the table.

        :param stream: the stream to write to
        :param cols: the column definition
        :param section_index: the index of the section, `0` for the first
            section
        :param n_rows: the number of rows that were written in the section
        """

    def begin_table_section_header(self, stream: TextIOBase, cols: str,
                                   section_index: int) -> None:
        """
        Begin the header row of a section.

        :param stream: the stream to write to
        :param cols: the column definition
        :param section_index: the index of the section, `0` for the first
            section
        """

    def end_table_section_header(self, stream: TextIOBase, cols: str,
                                 section_index: int) -> None:
        """
        End the header row of a section.

        :param stream: the stream to write to
        :param cols: the column definition
        :param section_index: the index of the section, `0` for the first
            section
        """

    def begin_table_row(self, stream: TextIOBase, cols: str,
                        section_index: int, row_index: int) -> None:
        """
        Begin a table header row, section row, or normal row in a section.

        :param stream: the stream to write to
        :param cols: the column definition
        :param section_index: the index of the current section, `-1` if we
            are in the table header
        :param row_index: the row index in the section: `0` for the first
            actual row, `-1` for a section header row, `-2` for a table
            header row
        """

    def end_table_row(self, stream: TextIOBase, cols: str,
                      section_index: int, row_index: int) -> None:
        """
        End a table header row, section row, or normal row in a section.

        :param stream: the stream to write to
        :param cols: the column definition
        :param section_index: the index of the current section
        :param row_index: the row index in the section: `0` for the first
            actual row, `-1` for a section header row, `-2` for a table
            header row
        """

    def begin_table_cell(self, stream: TextIOBase, cols: str,
                         section_index: int, row_index: int,
                         col_index: int) -> None:
        """
        Begin a header cell, section header cell, or normal cell.

        :param stream: the stream to write to
        :param cols: the column definitions
        :param section_index: the index of the current section, `-1` if this
            is a table header cell
        :param row_index: the row index in the section: `0` for the first
            actual row, `-1` for a section header row, `-2` for a table
            header row
        :param col_index: the column index, `0` for the first column
        """

    def end_table_cell(self, stream: TextIOBase, cols: str,
                       section_index: int, row_index: int,
                       col_index: int) -> None:
        """
        End a header cell, section header cell, or normal cell.

        :param stream: the stream to write to
        :param cols: the column definitions
        :param section_index: the index of the current section, `-1` if this
            is a table header cell
        :param row_index: the row index in the section: `0` for the first
            actual row, `-1` for a section header row, `-2` for a table
            header row
        :param col_index: the column index, `0` for the first column
        """

    def text(self, stream: TextIOBase, text: str, bold: bool, italic: bool,
             code: bool) -> None:
        """
        Write a chunk of text.

        :param stream: the stream to write to
        :param text: the text to write
        :param bold: is the text in bold face?
        :param italic: is the text in italic face?
        :param code: is the text in code face?
        """

    def render_numeric_exponent(self, e: str) -> str:
        """
        Render the exponent of a number.

        This function is for use in conjunction with
            :func:`moptipy.utils.strings.numbers_to_strings`.

        :param e: the exponent
        :returns: a rendered string
        """

    def filename(self,
                 file_name: str = "file",
                 dir_name: str = ".",
                 use_lang: bool = True) -> Path:
        """
        Get the right filename for this text driver.

        :param file_name: the base file name
        :param dir_name: the base directory
        :param use_lang: should we use the language to define the filename?
        :returns: the path to the file to generate
        """
        if not isinstance(dir_name, str):
            raise type_error(dir_name, "dir_name", str)
        if len(dir_name) <= 0:
            raise ValueError(f"invalid dir_name: '{dir_name}'.")
        if not isinstance(file_name, str):
            raise type_error(file_name, "file_name", str)
        if len(file_name) <= 0:
            raise ValueError(f"invalid file_name: '{file_name}'.")
        if not isinstance(use_lang, bool):
            raise type_error(use_lang, "use_lang", bool)
        out_dir = Path.directory(dir_name)
        suffix = str(self)
        if not isinstance(suffix, str):
            raise type_error(suffix, "result of str(table driver)", str)
        if len(suffix) <= 0:
            raise ValueError(f"invalid driver suffix: '{suffix}'")
        if use_lang:
            file_name = Lang.current().filename(file_name)
        file: Final[Path] = out_dir.resolve_inside(f"{file_name}.{suffix}")
        file.ensure_file_exists()
        return file



"""Classes for printing tables in a text format."""

from contextlib import AbstractContextManager
from io import TextIOBase
from typing import Final, Optional, Iterable, Union, Callable, List

from moptipy.utils.text_format import TextFormatDriver, FormattedStr
from moptipy.utils.types import type_error


class Table(AbstractContextManager):
    """
    The table context.

    This class provides a simple and hierarchically structured way to write
    tables in different formats. It only supports the most rudimentary
    formatting and nothing fancy (such as references, etc.). However, it may
    be totally enough to quickly produce tables with results of experiments.

    Every table must have a table header (see :meth:`header`).
    Every table then consists of a sequence of one or multiple sections
    (see :meth:`section` and :class:`Section`).
    Each table section itself may or may not have a header
    (see :meth:`Section.header`) and must have at least one row (see
    :meth:`Section.row` and :class:`Row`).
    Each row must have the exact right number of cells (see :meth:`Row.cell`).
    """

    def __init__(self, stream: TextIOBase, cols: str,
                 driver: TextFormatDriver):
        """
        Initialize the table context.

        :param stream: the stream to which all output is written
        :param cols: the columns of the table
        :param driver: the table driver
        """
        super().__init__()
        if not isinstance(stream, TextIOBase):
            raise type_error(stream, "stream", TextIOBase)
        if not isinstance(cols, str):
            raise type_error(cols, "cols", str)
        if not isinstance(driver, TextFormatDriver):
            raise type_error(driver, "driver", TextFormatDriver)

        cols = cols.strip()
        if len(cols) <= 0:
            raise ValueError(
                "cols must not be empty to just composed of white space, "
                f"but is '{cols}'.")
        for c in cols:
            if c not in ("c", "l", "r"):
                raise ValueError("each col must be c, l, or r, but "
                                 f"encountered {c} in {cols}.")
        #: the internal stream
        self.__stream: TextIOBase = stream
        #: the internal column definition
        self.__cols: Final[str] = cols
        #: the internal table driver
        self.__driver: Final[TextFormatDriver] = driver
        #: the header state: 0=no header, 1=in header, 2=after header
        self.__header_state: int = 0
        #: the section index
        self.__section_index: int = 0
        #: the row index
        self.__row_index: int = 0
        #: the column index
        self.__col_index: int = 0
        #: the section state: 0 outside of section, 1 inside of section,
        #: 2 after section
        self.__section_state: int = 0
        #: the section header state: 0=no header, 1=in header, 2=after header
        self.__section_header_state: int = 0
        #: the row state: 0=before row, 1=in row, 2=after row
        self.__row_state: int = 0

    def _begin_section(self):
        """Start a section."""
        if self.__stream is None:
            raise ValueError("table already closed, cannot start section.")
        if self.__header_state <= 0:
            raise ValueError("cannot start section before table header.")
        if self.__header_state <= 1:
            raise ValueError("cannot start section inside table header.")
        if self.__section_state == 1:
            raise ValueError("cannot start section inside section.")
        if self.__section_header_state == 1:
            raise ValueError("cannot start section inside section header.")
        if self.__section_header_state >= 1:
            raise ValueError(
                "cannot start section inside section after header.")
        if self.__row_state == 1:
            raise ValueError("cannot start section inside row.")
        self.__driver.begin_table_section(self.__stream, self.__cols,
                                          self.__section_index)
        self.__row_index = 0
        self.__section_state = 1
        self.__row_state = 0

    def _end_section(self):
        """End a section."""
        if self.__stream is None:
            raise ValueError("table already closed, cannot end section.")
        if self.__header_state <= 0:
            raise ValueError("cannot end section before table header.")
        if self.__header_state <= 1:
            raise ValueError("cannot end section inside table header.")
        if self.__section_state <= 0:
            raise ValueError("cannot start section outside section.")
        if self.__section_state >= 2:
            raise ValueError("cannot end section after section.")
        if self.__section_header_state == 1:
            raise ValueError("cannot end section inside section header.")
        if self.__row_index <= 0:
            raise ValueError(
                "cannot end section before writing any row.")
        if self.__row_state <= 0:
            raise ValueError("cannot end a section before a row.")
        if self.__row_state <= 1:
            raise ValueError("cannot end section inside of row.")
        self.__driver.end_table_section(
            self.__stream, self.__cols, self.__section_index,
            #  had_header is not used: self.__section_header_state == 2,
            self.__row_index)
        self.__row_index = 0
        self.__section_state = 2
        self.__section_header_state = 0
        self.__section_index += 1

    def _begin_row(self, header_mode: int):
        """
        Start a row.

        :param header_mode: the header mode: -2 table header, -1 section
            header, 0 normal row
        """
        if self.__stream is None:
            raise ValueError("table already closed, cannot start row.")
        if self.__header_state == 1:
            raise ValueError(
                "cannot start row inside table header.")
        if self.__section_header_state == 1:
            raise ValueError(
                "cannot start row inside section header.")
        if self.__section_state >= 2:
            raise ValueError(
                "cannot start row after section.")
        if self.__row_state == 1:
            raise ValueError("cannot start row inside row.")

        row_index: int
        sec_index: int

        if header_mode == -2:  # start table header
            if self.__header_state >= 2:
                raise ValueError(
                    "cannot start table header row after table header.")
            if self.__section_state == 1:
                raise ValueError(
                    "cannot start table header row inside section.")
            if self.__section_header_state >= 2:
                raise ValueError(
                    "cannot start table header row after section header.")
            if self.__row_state >= 2:
                raise ValueError("cannot start table header after row.")
            self.__header_state = 1
            self.__driver.begin_table_header(self.__stream, self.__cols)
            row_index = -2
            sec_index = -1

        elif header_mode == -1:  # start section header
            if self.__header_state < 2:
                raise ValueError(
                    "can only start section header row after table header.")
            if self.__section_header_state >= 2:
                raise ValueError(
                    "cannot start section header row after section header.")
            if self.__section_state <= 0:
                raise ValueError(
                    "cannot start section header row outside of section.")
            if self.__row_index > 0:
                raise ValueError("cannot start section header row "
                                 f"after {self.__row_index} rows.")
            if self.__row_state >= 2:
                raise ValueError("cannot start section header after row.")
            self.__section_header_state = 1
            sec_index = self.__section_index
            self.__driver.begin_table_section_header(
                self.__stream, self.__cols, sec_index)
            row_index = -1

        elif header_mode == 0:
            if self.__header_state < 2:
                raise ValueError("can only start row after table header.")
            if self.__section_state <= 0:
                raise ValueError("cannot start row outside of section.")
            sec_index = self.__section_index
            row_index = self.__row_index
        else:
            raise ValueError(f"invalid header mode {header_mode}.")

        self.__col_index = 0
        self.__row_state = 1
        self.__driver.begin_table_row(
            self.__stream, self.__cols, sec_index, row_index)

    def _end_row(self, header_mode: int):
        """
        End a row.

        :param header_mode: the header mode: -2 table header, -1 section
            header, 0 normal row
        """
        if self.__stream is None:
            raise ValueError("table already closed, cannot start row.")
        if self.__header_state == 0:
            raise ValueError(
                "cannot end row before table header.")
        if self.__section_state >= 2:
            raise ValueError("cannot end row after section.")
        if self.__row_state <= 0:
            raise ValueError("cannot end row before row begins.")
        if self.__row_state >= 2:
            raise ValueError("cannot end row after row has already ended.")
        if self.__col_index != len(self.__cols):
            raise ValueError(
                f"cannot end row after {self.__col_index} columns for table "
                f"with column definition {self.__cols}.")

        row_index: int
        sec_index: int

        if header_mode == -2:  # end table header
            if self.__header_state >= 2:
                raise ValueError(
                    "cannot end table header row after table header.")
            if self.__section_state == 1:
                raise ValueError(
                    "cannot end table header row inside section.")
            if self.__section_header_state == 1:
                raise ValueError(
                    "cannot end table header row inside section header.")
            if self.__section_header_state >= 2:
                raise ValueError(
                    "cannot end table header row after section header.")
            self.__header_state = 2
            row_index = -2
            sec_index = -1

        elif header_mode == -1:  # end section header
            if self.__header_state < 2:
                raise ValueError(
                    "can only end section header row after table header.")
            if self.__section_header_state <= 0:
                raise ValueError(
                    "cannot end section header row before section header.")
            if self.__section_header_state >= 2:
                raise ValueError(
                    "cannot end section header row after section header.")
            if self.__section_state <= 0:
                raise ValueError(
                    "cannot start section header row outside of section.")
            if self.__row_index > 0:
                raise ValueError("cannot end section header row "
                                 f"after {self.__row_index} rows.")
            self.__section_header_state = 2
            sec_index = self.__section_index
            row_index = -1

        elif header_mode == 0:
            if self.__header_state < 2:
                raise ValueError("can only end row after table header.")
            if self.__section_state <= 0:
                raise ValueError("cannot end row outside of section.")
            sec_index = self.__section_index
            row_index = self.__row_index
            self.__row_index += 1
        else:
            raise ValueError(f"invalid header mode {header_mode}.")

        self.__col_index = 0
        self.__row_state = 2
        self.__driver.end_table_row(self.__stream, self.__cols,
                                    sec_index, row_index)
        if header_mode == -2:
            self.__driver.end_table_header(self.__stream, self.__cols)
        elif header_mode == -1:
            self.__driver.end_table_section_header(
                self.__stream, self.__cols, sec_index)

    def _cell(self, text: Optional[Union[str, Iterable[str]]]):
        """
        Render a cell.

        :param text: the text to write
        """
        if self.__stream is None:
            raise ValueError("table already closed, cannot start row.")
        if self.__header_state == 0:
            raise ValueError(
                "cannot have a cell before the table header starts.")
        if self.__section_state >= 2:
            raise ValueError(
                "cannot have cell after section end.")
        col_index: Final[int] = self.__col_index
        if col_index >= len(self.__cols):
            raise ValueError(
                f"cannot add cell after {col_index} columns for table "
                f"with column definition {self.__cols}.")
        if self.__row_state <= 0:
            raise ValueError("cannot begin cell before beginning a row.")
        if self.__row_state >= 2:
            raise ValueError("cannot begin cell after end of a row.")

        row_index: int
        section_index: int = self.__section_index
        if self.__header_state == 1:
            row_index = -2
            section_index = -1
        else:
            if self.__section_state <= 0:
                raise ValueError(
                    "cannot begin cell after header outside of section.")
            if self.__section_header_state == 1:
                row_index = -1
            else:
                row_index = self.__row_index
        self.__col_index = col_index + 1

        self.__driver.begin_table_cell(self.__stream, self.__cols,
                                       section_index, row_index, col_index)

        def __printit(st, strm: TextIOBase = self.__stream,
                      wrt: Callable[[TextIOBase, str, bool,
                                     bool, bool], None] = self.__driver.text) \
                -> None:
            if st is None:
                return
            if isinstance(st, str):
                if isinstance(st, FormattedStr):
                    wrt(strm, st, st.bold, st.italic, st.code)
                else:
                    wrt(strm, st, False, False, False)
            elif isinstance(st, Iterable):
                for ss in st:
                    __printit(ss)
            else:
                raise type_error(st, "text", (Iterable, str))
        __printit(text)
        self.__driver.end_table_cell(
            self.__stream, self.__cols, section_index, row_index, col_index)

    def header(self) -> 'Row':
        """
        Construct the header of the table.

        :returns: a new managed header row
        """
        return Row(self, -2)

    def header_row(self, cells: Iterable[str]) -> None:
        """Print the header row with a single call."""
        if not isinstance(cells, Iterable):
            raise type_error(cells, "cells", Iterable)
        with self.header() as row:
            for cell in cells:
                row.cell(cell)

    def section(self) -> 'Section':
        """
        Create a new section of rows.

        :returns: a new managed row section
        """
        return Section(self)

    def section_cols(self, cols: List[List[Optional[str]]],
                     header_row: Optional[Iterable[Optional[str]]] = None):
        """
        Print a section columns-by-column.

        :param cols: an array which contains one list per column of the table.
        :param header_row: an optional header row
        """
        if not isinstance(cols, list):
            raise type_error(cols, "cols", list)
        if len(cols) != len(self.__cols):
            raise ValueError(
                f"expected {len(self.__cols)} columns ({self.__cols}), "
                f"but cols has length {len(cols)}.")
        max_rows = max(len(col) for col in cols)
        if max_rows <= 0:
            raise ValueError("There are no rows in the cols array?")

        with self.section() as sec:
            if header_row is not None:
                if not isinstance(header_row, Iterable):
                    raise type_error(header_row, "section header_row",
                                     Iterable)
                with sec.header() as head:
                    for cell in header_row:
                        head.cell(cell)

            for rowi in range(max_rows):
                with sec.row() as row:
                    for col in cols:
                        row.cell(None if rowi >= len(col) else col[rowi])

    def __enter__(self):
        """
        Enter the table in a `with` statement.

        :return: `self`
        """
        if self.__stream is None:
            raise ValueError("Table writing already finished!")
        self.__driver.begin_table_body(self.__stream, self.__cols)
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        """
        Close the table after leaving the `with` statement.

        :param exception_type: ignored
        :param exception_value: ignored
        :param traceback: ignored
        :returns: `True` to suppress an exception, `False` to rethrow it
        """
        if not (self.__stream is None):
            self.__driver.end_table_body(self.__stream, self.__cols)
            self.__stream = None
        if self.__section_state <= 0:
            raise ValueError("cannot end table before any section")
        if self.__section_state <= 1:
            raise ValueError("cannot end table inside a section")
        if self.__header_state <= 0:
            raise ValueError("cannot end table before table header")
        if self.__header_state <= 1:
            raise ValueError("cannot end table inside table header")
        if self.__section_header_state == 1:
            raise ValueError("cannot end table inside section header")
        return exception_type is None


class Section(AbstractContextManager):
    """A table section is a group of rows, potentially with a header."""

    def __init__(self, owner: Table):
        """
        Initialize the row section.

        :param owner: the owning table
        """
        if not isinstance(owner, Table):
            raise type_error(owner, "owner", Table)
        #: the owner
        self.__owner: Final[Table] = owner

    def __enter__(self):
        """
        Enter the row section in a `with` statement.

        :return: `self`
        """
        # noinspection PyProtectedMember
        self.__owner._begin_section()
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        """
        Close the row section after leaving the `with` statement.

        :param exception_type: ignored
        :param exception_value: ignored
        :param traceback: ignored
        :returns: `True` to suppress an exception, `False` to rethrow it
        """
        # noinspection PyProtectedMember
        self.__owner._end_section()
        return exception_type is None

    def header(self) -> 'Row':
        """
        Print the section header.

        :return: the header row
        """
        return Row(self.__owner, -1)

    def row(self) -> 'Row':
        """
        Create a row.

        :return: the new row
        """
        return Row(self.__owner, 0)


class Row(AbstractContextManager):
    """A row class."""

    def __init__(self, owner: Table, header_mode: int):
        """
        Initialize the row.

        :param owner: the owning table
        :param header_mode: the header mode
        """
        if not isinstance(owner, Table):
            raise type_error(owner, "owner", Table)
        if not isinstance(header_mode, int):
            raise type_error(header_mode, "header_mode", int)
        if not -2 <= header_mode <= 0:
            raise ValueError(f"Invalid header mode {header_mode}.")
        #: the owner
        self.__owner: Final[Table] = owner
        #: the header mode
        self.__header_mode: Final[int] = header_mode

    def cell(self, text: Optional[Union[str, Iterable[str]]] = None) -> None:
        """
        Render the text of a cell.

        As parameter `text`, you can provide either a string or a sequence of
        strings. You can also provide an instance of
        :class:`moptipy.utils.text_format.FormattedStr` or a sequence thereof.
        This allows you to render formatted text in a natural fashion.

        :param text: the text to write
        """
        # noinspection PyProtectedMember
        self.__owner._cell(text)

    def __enter__(self):
        """
        Enter the row in a `with` statement.

        :return: `self`
        """
        # noinspection PyProtectedMember
        self.__owner._begin_row(self.__header_mode)
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        """
        Close the row after leaving the `with` statement.

        :param exception_type: ignored
        :param exception_value: ignored
        :param traceback: ignored
        :returns: `True` to suppress an exception, `False` to rethrow it
        """
        # noinspection PyProtectedMember
        self.__owner._end_row(self.__header_mode)
        return exception_type is None
